/* Copyright (C) 2023 FIAS Frankfurt Institute for Advanced Studies, Frankfurt / Main
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Felix Weiglhofer [committer] */
#include "Reco.h"

#include <Monitor.hpp>
#include <System.hpp>

#include <xpu/host.h>

#include "config/Yaml.h"
#include "log.hpp"
#include "util/TimingsFormat.h"
#include "util/TsUtils.h"

using namespace cbm::algo;

Reco::Reco() {}
Reco::~Reco() {}

void Reco::Validate(const Options& opts)
{
  if (!fs::exists(opts.ParamsDir())) {
    throw std::runtime_error("ParamsDir does not exist: " + opts.ParamsDir().string());
  }

  for (auto detector : opts.Detectors()) {
    switch (detector) {
      case fles::Subsystem::STS: break;
      default: throw std::runtime_error(fmt::format("Unsupported system: {}", ToString(detector)));
    }
  }
}

void Reco::Init(const Options& opts)
{
  if (fInitialized) throw std::runtime_error("Chain already initialized");

  Validate(opts);

  fContext.opts = opts;
  SetContext(&fContext);
  fUnpack.SetContext(&fContext);
  fStsHitFinder.SetContext(&fContext);

  xpu::device_prop props {xpu::device::active()};
  L_(info) << "Running CBM Reco on Device " << props.name();

  if (!opts.MonitorUri().empty()) {
    fContext.monitor = std::make_unique<cbm::Monitor>(opts.MonitorUri());
    L_(info) << "Monitoring enabled, sending to " << opts.MonitorUri();
  }

  // Reco Params
  fs::path recoParamsPath = opts.ParamsDir() / "RecoParams.yaml";
  YAML::Node yaml         = YAML::LoadFile(recoParamsPath.string());
  fContext.recoParams     = config::Read<RecoParams>(yaml);

  // Unpackers
  fUnpack.Init();

  // STS Hitfinder
  fs::path stsHitfinderParamsPath    = opts.ParamsDir() / "StsHitfinder.yaml";
  yaml                               = YAML::LoadFile(stsHitfinderParamsPath.string());
  sts::HitfinderPars hitFinderConfig = config::Read<sts::HitfinderPars>(yaml);
  hitFinderConfig.landauTable        = sts::LandauTable::FromFile(opts.ParamsDir() / "LandauWidthTable.txt");
  fStsHitFinder.SetParameters(hitFinderConfig);

  fInitialized = true;

  xpu::push_timer("Reco");
}

void Reco::Run(const fles::Timeslice& ts)
{
  if (!fInitialized) { throw std::runtime_error("Chain not initialized"); }

  for (uint64_t comp = 0; comp < ts.num_components(); comp++) {
    xpu::t_add_bytes(ts.size_component(comp));
  }

  xpu::push_timer(fmt::format("TS {}", ts.index()));

  L_(info) << ">>> Processing TS " << ts.index();
  xpu::set<cbm::algo::Params>(Params());

  std::vector<CbmStsDigi> digis;
  switch (Params().sts.unpackMode) {
    case RecoParams::UnpackMode::XPU:
      // digis = fUnpackXpu.Exec(ts);
      throw std::runtime_error("XPU unpacker currently not implemented");
      break;
    default:
    case RecoParams::UnpackMode::CPU: digis = fUnpack.Run(ts).first.fData.fSts.fDigis; break;
  }
  fStsHitFinder(digis);

  xpu::timings ts_times = xpu::pop_timer();

  if (HasMonitor())
    GetMonitor().QueueMetric("cbm_reco", {{"hostname", fles::system::current_hostname()}},
                             {{"tsProcessed", 1}, {"bytesProcessed", ts_utils::SizeBytes(ts)}});


  PrintTimings(ts_times);
}

void Reco::Finalize()
{
  fStsHitFinder.Finalize();

  // Pop timer that was started in Init()
  xpu::timings t = xpu::pop_timer();
  if (Opts().CollectKernelTimes()) {
    L_(info) << MakeReportSubtimers("Run Summary", fTimesliceTimesAcc) << "\n" << MakeReportSummary("Total", t);
  }
  else {
    L_(info) << "Total Processing time (Wall): " << t.wall() << " ms";
  }
}

void Reco::PrintTimings(xpu::timings& timings)
{
  if (Opts().CollectKernelTimes()) {
    L_(info) << MakeReportSubtimers("TS timings", timings) << "\n" << MakeReportSummary("Total", timings);
    fTimesliceTimesAcc.merge(timings);
  }
  else {
    L_(info) << "TS Processing time (Wall): " << timings.wall() << " ms";
  }
}
